#!/usr/bin/env python

#
# (c) 2015-2020 caregraf
#

import urllib
import urllib2
import ssl
import json
import re
from StringIO import StringIO
from collections import OrderedDict
import logging

from rpcutils.brokerRPC import VistARPCConnection, RPCLogger

"""
FMQL Endpoint Indirection - broker or REST and REST can be CSP (which has quirky errors)

TODO: merge in CacheInterface (may distinguish 3 - CSP/REST/BrokerIF
"""
                                
class FMQLRESTIF:
    """
    Indirection between endpoint for query so can go to CSPs etc.
        
    TODO: catch 503 etc in here and return clean ERROR codes
    """
    def __init__(self, fmqlEP, epWord="query"):
        self.fmqlEP = fmqlEP
        self.epWord = epWord
        
    def __str__(self):
        return "ENDPOINT: " + self.fmqlEP
        
    def invokeQuery(self, query):
        queryURL = self.fmqlEP + "?" + urllib.urlencode({self.epWord: query}) 
        try:
            if re.match("https", queryURL):
                # ignore bad cert - happens on test systems.
                ctx = ssl.create_default_context()
                ctx.check_hostname = False
                ctx.verify_mode = ssl.CERT_NONE
                sreply = StringIO(urllib2.urlopen(queryURL, context=ctx).read())
            else:
                sreply = StringIO(urllib2.urlopen(queryURL).read()) # be compatible with cacheObjectInterface
            # Want to preserve order of keys (as FMQL does) 
            jreply = json.load(sreply, object_pairs_hook=OrderedDict) 
            return jreply
        except:
            try:
                sreply.seek(0)
            except:
                logging.error("Problem Response for query {} - can't even examine".format(query))
                raise
            text = sreply.read()
            if re.search(r'FMQL\.CSP', self.fmqlEP): # CSP endpoint
                #
                # Error: <b>Cannot allocate a license</b><br>
                # ErrorNo: <b>5915</b><br>
                # CSP Page: <b>/csp/fmquery/FMQL.CSP</b><br>
                # Namespace: <b>CHCS</b><br>
                #
                if re.search(r'ErrorNo', text):
                    error = re.search(r'Error: \<b\>([^\<]+)', text).group(1)
                    errorNo = re.search(r'ErrorNo: \<b\>([^\<]+)', text).group(1)
                    logging.error("CSP Error Response for query {} - {} - {}".format(query, errorNo, error))
            last500 = text[-500:]
            if len(text) > 500:
                last500 = "[LAST 500] {}".format(last500)
            logging.error("Problem Response for query {} - {}".format(query, last500))
            raise
            
class FMQLBrokerIF:

    def __init__(self, hostname, port, access, verify, osehraVISTA=False):
        conn = VistARPCConnection(hostname, int(port), access, verify, "CG FMQL QP USER", RPCLogger(), useOSEHRACipher=osehraVISTA)
        conn.connect()
        self.__connection = conn
        self.__epDescr = hostname + ":" + str(port)
        
    def __str__(self):
        return "BROKER ENDPOINT: " + self.__epDescr
        
    def invokeQuery(self, query):
    
        reply = self.__connection.invokeRPC("CG FMQL QP", [query])

        try: 
            # preserve order as FMQL does
            jreply = json.loads(reply, object_pairs_hook=OrderedDict)
        except:
            logging.error("Can't make JSON for response to query {} - {}".format(query, reply))
            raise
        
        return jreply



#!/usr/bin/env python
# -*- coding: utf8 -*-

#
# (c) 2015-2020 caregraf
#

import os
import re 
import sys
import json
import shutil
from datetime import datetime
from collections import OrderedDict, defaultdict
import logging
from fmqlIF import FMQLRESTIF, FMQLBrokerIF
from cacherUtils import FMQLReplyStore, cacheFileToStore, ensureCacheLocations, configLogging, DATAV1_LOCN_TEMPL, SCHEMA_LOCN_TEMPL, LOG_LOCN_TEMPL

"""
Basic FileMan data and schema cacher using FMQL.

TODO: FMQLSchemaStore to parallel FMQLReplyStore (may become FMQLDataStore)

Background: FileMan doesn't have enough indexes to support analytics. To fully analyze
data must first be cached. It can then be processed in a different DB (ex/ Mongo) or just
from disk.

Cache to "/data/vista/{stationNumber}/DataV1/" which it will create if it is absent

To test:
- du -h /data/{SYSTEM}/Data[V2] and /Schema
- old 2 download has 'File 2, limit 500 : all 1032150 of file 2 in 4:49:22.820342'

Invoke: 

    > nohup python -u cacher.py VISTACONFIG > nohupVISTA.out & and tail -f nohupVISTA.out; monitor with ps -aux, kill with kill -9 {pid}

or 'python cacher.py VISTACONFIG [SCHEMA]'

Note: all functions here are driven from Cacher config
"""

# ########################## Data #########################

def cacheData(config):

    start = datetime.now()
        
    # Note: can raise Exception if can't create/use configured locations - will be caught in invoker 
    ensureCacheLocations(config["stationNumber"])

    configLogging(config["stationNumber"], "cacherLog", config["logLevel"] if "logLevel" in config else "INFO")
    logging.info("caching in {}".format(DATAV1_LOCN_TEMPL.format(config["stationNumber"])))

    systemId = config["stationNumber"] if "name" not in config else "{} [{}]".format(config["name"], config["stationNumber"])
    logging.info("Caching Data of VistA {} [PID {}]".format(systemId, os.getpid()))
    
    fmqlIF = makeFMQLIF(config)
    
    useZip = config["useZip"] if "useZip" in config else True
    # makeDir False as explicitly building below (may change later)
    fmqlReplyStore = FMQLReplyStore(DATAV1_LOCN_TEMPL.format(config["stationNumber"]), isOrdered=True, useZip=useZip, makeDir=False)
    
    trackZeroFile = "{}trackZeroFiles{}.json".format(LOG_LOCN_TEMPL.format(config["stationNumber"]), config["stationNumber"])
    try:
        zeroFiles = json.load(open(trackZeroFile))
    except: 
        zeroFiles = []
        
    excludeTypes = set(config["excluded"]) if "excluded" in config else set()
    includeTypes = set(config["included"]) if "included" in config else None

    #
    # Option of Count Order or Id Order
    #
    # ... may retire: ala FMQLReplyStore, support a Schema Store
    # 
    def filesFromSchema(fmqlEP, schemaCacheLocation, allFiles=False, typeOrder=False):
        """
        From SELECT TYPES - if in cache just load and read. Otherwise cache and
        then load and read.
        """
        query = "SELECT TYPES TOPONLY" # doing TOPONLY and not POPONLY in case POP is wrong
        cacheFile = re.sub(r' ', "_", query) + ".json"
        try:
            reply = json.load(open(schemaCacheLocation + cacheFile), object_pairs_hook=OrderedDict)
        except Exception:
            logging.info("First time through - must (re)cache {} to {} ...".format(query, schemaCacheLocation))
            reply = fmqlEP.invokeQuery(query)
            json.dump(reply, open(schemaCacheLocation + cacheFile, "w"))
        if typeOrder: # used if included explicit - otherwise order biggest first
            oresults = sorted(reply["results"], key=lambda res: float(re.sub(r'\_', '.', res["number"])))
        else:      
            oresults = sorted(reply["results"], key=lambda res: int(res["count"]) if "count" in res else 0, reverse=True)
        filesInOrder = []
        for i, result in enumerate(oresults, 1):
            fileId = re.sub(r'\.', '_', result["number"])
            filesInOrder.append({"id": fileId, "count": 0 if "count" not in result else int(result["count"])})
        logging.info("Returning {} top files - not all will have data".format(len(filesInOrder)))
        return filesInOrder

    j = 0
    defaultLimit = 5000 if "defaultLimit" not in config else config["defaultLimit"]
    defaultCStop = 1000 if "defaultCStop" not in config else config["defaultCStop"]
    problem = False
    typeOrder = True if includeTypes else False # cache in type order if list is explicit
    schemaCacheLocation = SCHEMA_LOCN_TEMPL.format(config["stationNumber"])
    for i, fli in enumerate(filesFromSchema(fmqlIF, schemaCacheLocation, allFiles=False, typeOrder=typeOrder), 1):

        if fli["id"] in excludeTypes:
            logging.debug("{}. File {} expecting count of {} - EXCLUDED SO SKIPPING".format(i, fli["id"], fli["count"]))
            continue

        if fli["id"] in zeroFiles:
            logging.debug("{}. File {} expecting count of {} - KNOWN ZERO SO SKIPPING".format(i, fli["id"], fli["count"]))
            continue

        if includeTypes and fli["id"] not in includeTypes:
            logging.debug("{}. File {} expecting count of {} - NOT IN EXPLICIT INCLUDE LIST SO SKIPPING".format(i, fli["id"], fli["count"]))
            continue

        logging.info("{}. File {} expecting count of {}{}".format(i, fli["id"], fli["count"], " ** Note POP says 0 but trying anyhow" if fli["count"] == 0 else ""))
            
        if "explicitLimits" in config and fli["id"] in config["explicitLimits"]:
            limit = config["explicitLimits"][fli["id"]]
        else:
            limit = defaultLimit

        if "explicitCStops" in config and fli["id"] in config["explicitCStops"]:
            cstop = config["explicitCStops"][fli["id"]]
        else:
            cstop = defaultCStop

        try: # may get license error - make sure save zero files first
            if not cacheFileToStore(fmqlIF, fmqlReplyStore, fli["id"], limit=limit, cstop=cstop, maxNumber=-1):
                zeroFiles.append(fli["id"])
        except Exception, e:
            logging.exception(e)
            problem = True
            break

    timeTaken = str(datetime.now() - start)
    if problem:
        logging.info("Caching finished UNSUCCESSFULLY at {} and took {}".format(datetime.now(), timeTaken))
    else:
        logging.info("Caching finished successfully at {} and took {}".format(datetime.now(), timeTaken))
    json.dump(zeroFiles, open(trackZeroFile, "w"))
    
# ####################### Schema #####################
    
def cacheSchemas(config):

    start = datetime.now()
    
    ensureCacheLocations(config["stationNumber"])
    configLogging(config["stationNumber"], "cacherLog", config["logLevel"] if "logLevel" in config else "DEBUG")

    systemId = config["stationNumber"] if "name" not in config else "{} [{}]".format(config["name"], config["stationNumber"])
    logging.info("Caching Schema of VistA {} [PID {}]".format(systemId, os.getpid()))

    fmqlIF = makeFMQLIF(config)
    
    schemaCacheLocation = SCHEMA_LOCN_TEMPL.format(config["stationNumber"])
    query = "SELECT TYPES"
    try:
        jreply = json.load(open(schemaCacheLocation + "SELECT_TYPES.json"))
    except:
        logging.info("First time through - must (re)cache {} to {} ...".format(query, schemaCacheLocation))
        jreply = fmqlIF.invokeQuery(query)
        json.dump(jreply, open(schemaCacheLocation + "SELECT_TYPES.json", "w"))

    fileIds = [re.sub(r'\.', "_", result["number"]) for result in jreply["results"]]
    logging.info("Must cache schema of {} files".format(len(fileIds)))

    alreadyCached = [re.match(r'SCHEMA\_([^\.]+)', f).group(1) for f in os.listdir(schemaCacheLocation) if os.path.isfile(os.path.join(schemaCacheLocation, f)) and re.search('\.json$', f) and re.match("SCHEMA_", f)]
    for i, fileId in enumerate(fileIds, 1):
        if fileId in alreadyCached:
            logging.info("{}. Schema of {} Already Cached".format(i, fileId))
            continue
        logging.info("{}. Caching Schema of {}".format(i, fileId))
        query = "DESCRIBE TYPE " + fileId
        queryStart = datetime.now()
        jreply = fmqlIF.invokeQuery(query)
        json.dump(jreply, open(schemaCacheLocation + "SCHEMA_" + fileId + ".json", "w"))

    # Phase 2 - cache SELECT TYPE REFS
    alreadyCached = [re.match(r'REFS\_([^\.]+)', f).group(1) for f in os.listdir(schemaCacheLocation) if os.path.isfile(os.path.join(schemaCacheLocation, f)) and re.search('\.json$', f) and re.match("REFS\_", f)]
    for i, fileId in enumerate(fileIds, 1):
        if fileId in alreadyCached:
            logging.info("{}. Refs of {} Already Cached".format(i, fileId))
            continue
        logging.info("{}. Caching Refs of {}".format(i, fileId))
        query = "SELECT TYPE REFS " + fileId
        queryStart = datetime.now()
        jreply = fmqlIF.invokeQuery(query)
        json.dump(jreply, open(schemaCacheLocation + "REFS_" + fileId + ".json", "w"))
        
    timeTaken = str(datetime.now() - start)
    logging.info("Caching SCHEMA/REF finished successfully at {} for {} files and took {}".format(datetime.now(), len(fileIds), timeTaken))
    
# #################### FMQLIF wrapper (schema and data) #######################
    
def makeFMQLIF(config):
    # Should work for CSP and 'regular' REST endpoint
    if "fmqlEP" in config:
        fmqlIF = FMQLRESTIF(config["fmqlEP"], epWord=config["fmqlQuery"])
        logging.info("Using REST Interface {}".format(config["fmqlEP"]))
    else:
        if "broker" not in config:
            raise Exception("Exiting - invalid 'config': neither 'broker' nor REST ('fmqlEP') settings available")
        # Must a/c for cypher in OSEHRA being different than Cypher in regular production VISTA (or maybe not if Vagrant VISTA is changed!)
        osehraVISTA = config["broker"]["osehraVISTA"] if "osehraVISTA" in config["broker"] else False
        fmqlIF = FMQLBrokerIF(config["broker"]["hostname"], config["broker"]["port"], config["broker"]["access"], config["broker"]["verify"], osehraVISTA=osehraVISTA)
        logging.info("Using RPC Broker Interface {}:{}".format(config["broker"]["hostname"], config["broker"]["port"]))
    return fmqlIF

# ############################# Driver ####################################
    
# ./cacher.py {ConfigBase} [SCHEMA] where Config is in ConfigFile.json
def main():

    assert(sys.version_info >= (2,7))

    if len(sys.argv) < 2:
        print "need to specify configuration ex/ {VISTAAB}.json - exiting"
        return

    configBase = sys.argv[1].split(".")[0]

    if not os.path.isfile("{}.json".format(configBase)):
        print "No config file {}.json - exiting".format(configBase)
        return

    config = json.load(open("{}.json".format(configBase)), object_pairs_hook=OrderedDict)

    if "stationNumber" not in config:
        print "No station number in configuration - exiting"
        return

    try:
        if len(sys.argv) == 3:
            if sys.argv[2] == "SCHEMA":
                cacheSchemas(config)
            else:
                print "Invalid second argument {} when only SCHEMA|REPORT allowed - exiting".format(sys.argv[2])
        else:
            cacheData(config)
    except Exception, e:
        # common issue is base location doesn't exist or isn't writable
        print "Unable to proceed: {} - exiting".format(str(e))
        return

if __name__ == "__main__":
    main()




#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import sys
import json
import shutil
from datetime import datetime
import re
from collections import defaultdict, OrderedDict
import logging

from cacherUtils import FMQLReplyStore, V1RecordToV2, ensureCacheLocations, DATAV1_LOCN_TEMPL, DATA_LOCN_TEMPL, LOG_LOCN_TEMPL

"""
Turn FMQL v1.* form replies to v2 form 
- supports both zip and "raw" JSON caches, for both source and target
  - REM: v1 is zip or json; v2 can be either, set with 'useZip'
- will restart where it left off or rebuild from scratch
- notes any transformation errors (shouldn't happen)

TODO: once flipper in, allow to apply flipper AFTER doing transform basic
ie/ pipeline
"""
def toV2Data(stationNumber, onlyTypes=None, forceRedo=False, useZip=True, forMongo=True):

    ensureCacheLocations(stationNumber)

    log = logging.getLogger('')
    log.setLevel(logging.DEBUG)    
    ch = logging.StreamHandler(sys.stdout)
    format = logging.Formatter("%(levelname)s - %(message)s")
    ch.setFormatter(format)
    log.addHandler(ch)

    systemId = stationNumber
    logging.info("V2ing Data of VistA {} [PID {}]".format(systemId, os.getpid()))
    
    v1ReplyStore = FMQLReplyStore(DATAV1_LOCN_TEMPL.format(stationNumber), isOrdered=True)
    logging.info("Started Reply Cache V1 - uses {} - total replies {}".format("ZIP" if v1ReplyStore.usesZip() else "JSON", v1ReplyStore.totalReplies()))
        
    v2ReplyStore = FMQLReplyStore(DATA_LOCN_TEMPL.format(stationNumber), useZip=useZip)
    logging.info("Started Reply Cache V2 - uses {} - total replies {}".format("ZIP" if v2ReplyStore.usesZip() else "JSON", v2ReplyStore.totalReplies()))
    
    start = datetime.now()
    logging.info("Starting transformation of V1 to V2 Data at {}".format(str(start)))

    # May be .zip even if v1 is json
    existingReplyFiles = set(v2ReplyStore.replyFiles())
    def excludeV2AlreadyReply(replyFile):
        if replyFile in existingReplyFiles:
            logging.info("{} Being Skipped as already V2 (no forceRedo)".format(replyFile))
            return True
        return False
        
    replyIterator = v1ReplyStore.iterator(onlyTypes=onlyTypes, excludeRepliesFilter=excludeV2AlreadyReply if not forceRedo else None)
    problems = []
    overallWStops = {}
    ortn = V1RecordToV2(forMongo=forMongo)
    transformed = 0
    transformedRecords = 0
    startT = datetime.now() # start before iteration
    for reply in replyIterator.next():
        
        nreply = OrderedDict([("results", [])])
        for nrProp in reply:
            if nrProp in ["count", "results"]:
                continue
            nreply[nrProp] = reply[nrProp]
                
        for record in reply["results"]:
            try: 
                nrecord = ortn.transform(record)
                transformedRecords += 1
                nreply["results"].append(nrecord)
                if "fmqlHasStops" in nrecord:
                    if "fmqlHasStops" not in nreply:
                        nreply["fmqlHasStops"] = -1
                    if nreply["fmqlHasStops"] < nrecord["fmqlHasStops"]:
                        nreply["fmqlHasStops"] = nrecord["fmqlHasStops"]
            except Exception as e:
                logging.exception(e)
                logging.info("EXCEPTION transforming in {}".format(record["uri"]["value"]))
                problems.append(record["uri"]["value"])
                continue
        
        nreplyFile = v2ReplyStore.flush(nreply)
        # Want to know if results in reply embed stop information
        if "fmqlHasStops" in nreply:
            overallWStops[nreplyFile] = nreply["fmqlHasStops"]
        transformed += 1
        logging.info("{}. Transformed and Flushed {} in {} out of {}".format(transformed, nreplyFile, str(datetime.now() - startT), str(datetime.now() - start)))
        startT = datetime.now()
        
    timeTaken = str(datetime.now() - start)
    if len(overallWStops) == 0:
        logging.info("Ending transformation - {:,} records in {:,} replies, problem {}, NO STOPS, over {}".format(transformedRecords, transformed, len(problems), timeTaken))
    else:
        logging.info("Ending transformation - {:,} records in {:,} replies, problem {}, w/stop(s) {:,}, max stopped length {:,}, over {}".format(transformedRecords, transformed, len(problems), len(overallWStops), max(overallWStops[rf] for rf in overallWStops), timeTaken))

# ############################# Driver ####################################

"""
Form of configs

    {
        "stationNumber": "999",
        "included": ["2"],
        "forceRedo": False,
        "forMongo": True
    }
    
"""
def main():

    if len(sys.argv) < 2:
        print "need to specify configuration ex/ {VISTAAB}.json - exiting"
        return

    configBase = sys.argv[1].split(".")[0]

    if not os.path.isfile("{}.json".format(configBase)):
        print "No config file {}.json - exiting".format(configBase)
        return

    config = json.load(open("{}.json".format(configBase)), object_pairs_hook=OrderedDict)

    if "stationNumber" not in config:
        print "No station number in configuration - exiting"
        return
    stationNumber = config["stationNumber"]

    if "included" not in config or len(config["included"]) == 0:
        types = None
    else:
        types = config["included"]
        
    forceRedo = True if "forceRedo" in config and config["forceRedo"] == True else False
    forMongo = False if "forMongo" in config and config["forMongo"] == False else True
    
    toV2Data(stationNumber, onlyTypes=types, forceRedo=forceRedo, forMongo=forMongo)

if __name__ == "__main__":
    main()



#!/usr/bin/env python

#
# (c) 2015-2020 caregraf
#

import urllib, urllib2

class CacheObjectInterface:
    """
    Utility for talking to FMQL through a Cache Object Interface. Key feature is managing sessions. Cache uses cookies for session identification. Cache limits the number of sessions on a server so it is important to use and reuse the same session.
    
    If Cache runs out of sessions, it will issue Service Unavailable 503 errors.
    
    Follows FMQLInterface
    """
    def __init__(self, ep):
        # ex/ http://...../FMQL.csp
        self.ep = ep
        self.cookie = ""
        
    def invokeRPC(self, name, params):
        # to match broker if for now
        return self.invokeQuery(params[0]).read()
        
    def invokeQuery(self, query):
        queryDict = {"query": query}
        queryurl = self.ep + "?" + urllib.urlencode(queryDict)
        request = urllib2.Request(queryurl)
        if self.cookie:
            request.add_header('cookie', self.cookie)
        try:
            response = urllib2.urlopen(request)
        except urllib2.URLError, e:
            # 503 "Service Unavailable": The server cannot process the request due to a high load
            raise
        # Always reset the cookie - may be a new one if session idle for > 15 minutes.
        # SET-COOKIE: CSPSESSIONID-SP-57772-UP-csp-fmquery-=0010000100002g3gWldo9l0000fAj6SQgextDm2AmskX7GxQ--; path=/csp/fmquery/;  httpOnly;
        self.cookie = response.info().getheader('Set-Cookie')
        return response      
        



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import sys
import re
import json
from collections import OrderedDict, Counter
from datetime import datetime
import logging

from .. import VISTA_DATA_BASE_DIR
from ..cacher.cacherUtils import FMQLReplyStore, FilteredResultIterator, BASE_LOCN_TEMPL, DATA_LOCN_TEMPL
from reduceReportType import TypeReducer, SubTypeReducer, reportReductions, DEFAULT_FORCE_COUNT_PTER_TYPES, isValidDayForm

TYPER_LOCN_TEMPL = BASE_LOCN_TEMPL + "Typer/"
TYPER_REDUCTIONS_LOCN_TEMPL = TYPER_LOCN_TEMPL + "Reductions/"
TYPER_REPORTS_LOCN_TEMPL = TYPER_LOCN_TEMPL + "Reports/"

"""
As command, fmqltyper

Same pattern as v2er, cacher, using 
- [a] /data/vista/{STATIONNO}/... locations for reductions and reports and 
- [b] taking a configuration from the location the runtime runs from. 
- [c] that 'local' location is usually just a place for these configurations
- [d] onAndAfterDay/upToDay rely on create property and if set, want a "reductionLabel" to differentiate them

TODO:
- try yaml config https://pyyaml.org/wiki/PyYAMLDocumentation as can embed code 
easily ie/ include custom filters/enhancers inline so no need for custom ie/ can't
now invoke dynamically in command line
- for now, using print, not logging. May move over to one pattern 
"""

def reduceReportTypes(stationNumber, types, dataLocnTempl, forceRedo=False, createDate={}, reductionLabel="", subTypeProps={}, forceCountPointerTypesExtra={}, forceCountProperties={}, onAndAfterDay="", upToDay="", countDateMonth=False, countDateMonthFor=[], customEnhancers={}):

    start = datetime.now()
    
    log = logging.getLogger('')
    log.setLevel(logging.DEBUG)    
    ch = logging.StreamHandler(sys.stdout)
    format = logging.Formatter("%(levelname)s - %(message)s")
    ch.setFormatter(format)
    log.addHandler(ch)
    
    # Used to ensure 
    def ensureLocation(locn):
        if not os.path.isdir(locn):
            os.mkdir(locn)
        return locn
    ensureLocation(TYPER_REDUCTIONS_LOCN_TEMPL.format(stationNumber))
    ensureLocation(TYPER_REPORTS_LOCN_TEMPL.format(stationNumber))
                    
    typesAndDates = dict((typId, createDate[typId] if typId in createDate else "") for typId in types)
    
    if onAndAfterDay:
        if not isValidDayForm(onAndAfterDay):
            raise Exception("Invalid 'on and after day': {}".format(onAndAfterDay))

    if upToDay:
        if not isValidDayForm(upToDay):
            raise Exception("Invalid 'up to day': {}".format(upToDay))
            
    if (upToDay or onAndAfterDay) and reductionLabel == "":
        raise Exception("You must name the reduction ('reductionLabel') for time limited ('onAndAfterDay'/'upToDay') reductions")
        
    def getCreateDay(resource, createDateProp):
        value = resource[createDateProp]["value"] if "value" in resource[createDateProp] else resource[createDateProp]["label"]
        dayValue = value.split("T")[0] # using lex comp on ISO‑8601 form as fastest
        if not isValidDayForm(dayValue):
            log.debug("Can't get create day as its date has invalid value: {}".format(dayValue))
            return None
        return dayValue
            
    logging.info("About to reduce (and report) on {} types, starting at {} [PID {}]".format(len(typesAndDates), start, os.getpid()))

    count = 0 # allow for 0 - no i defined
    filtered = 0      
    dataLocn = dataLocnTempl.format(stationNumber)  
    for i, typId in enumerate(sorted(typesAndDates, key=lambda x: float(re.sub(r'\_', '.', x))), 1):
        startTyp = start if len(types) == 1 else datetime.now()
            
        reductionLocn = ensureLocation(TYPER_REDUCTIONS_LOCN_TEMPL.format(stationNumber))
        redFFL = "{}/{}{}Reduction.json".format(reductionLocn, typId, reductionLabel)
        
        if forceRedo == False and os.path.isfile(redFFL):
            logging.info("Not reducing (and reporting) {} as already exists and NOT forceredo".format(typId))
            continue

        logging.info("Reducing (and reporting) {}".format(typId))
        
        startAtReply = ""
        if onAndAfterDay:
            if not re.search(r'RF', dataLocn):
                store = FMQLReplyStore(dataLocn)
                startAtReply = store.firstReplyFileOnOrAfterCreateDay(typId, typesAndDates[typId], onAndAfterDay)
                if startAtReply == "":
                    logging.info("** Exiting: can't find Reply to start at on and after day {}, create property {}".format(onAndAfterDay, typesAndDates[typId]))
                    break
                logging.debug("Start Efficiency: configuring Iterator for on and after time: on or after day {}, create property {}, starting at reply {}".format(onAndAfterDay, typesAndDates[typId], startAtReply))
            else:
                logging.debug("No Start Efficiency: taking first Reply and ignoring onAndAfterDay as walking reformed (RF) data which is usually from out of order multiples")
        resourceIter = FilteredResultIterator(dataLocn, typId, isOrdered=False, startAtReply=startAtReply)
        fcpts = DEFAULT_FORCE_COUNT_PTER_TYPES
        if typId in forceCountPointerTypesExtra:
            fcpts.extend(forceCountPointerTypesExtra[typId])
        fcps = forceCountProperties[typId] if typId in forceCountProperties else []
        countDateMonthType = True if (countDateMonth or typId in countDateMonthFor) else False
        if typId in subTypeProps: 
            if isinstance(subTypeProps[typId], list):
                stps = subTypeProps[typId]
            else:
                stps = [subTypeProps[typId]]
            logging.info("Subtyping by '{}'".format("-".join(stps)))
            rtr = SubTypeReducer(typId, createDateProp=typesAndDates[typId], subTypeProps=stps, forceCountPointerTypes=fcpts, forceCountProperties=fcps, countDateMonth=countDateMonthType)
        else:
            rtr = TypeReducer(typId, createDateProp=typesAndDates[typId], forceCountPointerTypes=fcpts, forceCountProperties=fcps, countDateMonth=countDateMonthType)
        customEnhancer = None
        if typId in customEnhancers:
            customEnhancer = customEnhancers[typId]
            
        msgThres = 50000 
        upToDayStopped = False 
        filtered = Counter()  
        for i, resource in enumerate(resourceIter, 1):
            count += 1
            filteredSoFar = sum(filtered[c] for c in filtered)
            if i % msgThres == 0:
                logging.debug("Checked {} more resources for total of {:,}, sub type reducers {:,}, so far filtered {:,}, now in reply {} - {}".format(msgThres, i, rtr.subTypeReducerCount(), filteredSoFar, resourceIter.currentReplyFile(), datetime.now() - start))
            idProp = "id" if "id" in resource else "_id"
            if customEnhancer:
                resource = customEnhancer.enhance(resource)
                if not resource:
                    filtered["CUSTOM"] += 1
                    continue
            # Allows for out of order too
            if onAndAfterDay or upToDay:
                if typesAndDates[typId] not in resource:
                    filtered["NOCREATEDATE"] += 1
                    continue
                createDay = getCreateDay(resource, typesAndDates[typId])
                if not createDay:
                    filtered["BADCREATEDATE"] += 1
                    continue
                if onAndAfterDay and createDay < onAndAfterDay:
                    filtered["LTONAFTER"] += 1
                    continue
                # Can't efficiently cut out as start for appt can go on (ie/ not ordered)
                if upToDay and createDay >= upToDay:
                    filtered["GTEUPTO"] += 1
                    continue 
            rtr.transform(resource)
            
        pattData = rtr.reductions()
        for red in pattData:
            red["_stationNumber"] = stationNumber
            if reductionLabel:
                red["_label"] = reductionLabel
                            
        json.dump(pattData, open(redFFL, "w")) 
     
        reportLocn = ensureLocation(TYPER_REPORTS_LOCN_TEMPL.format(stationNumber))
        reportReductions(pattData, "{}/{}{}Report.txt".format(reportLocn, typId, reductionLabel))
        
        if len(filtered):
            filteredMU = "/".join(["{} [{}]".format(reason, filtered[reason]) for reason in filtered])
            logging.info("Finished after {:,} resources of {} in {} at {}, {:,} explicitly filtered with {}".format(count, typId, datetime.now() - startTyp, datetime.now(), sum(filtered[c] for c in filtered), filteredMU))
        else:
            logging.info("Finished after {:,} resources of {} in {} at {}".format(count, typId, datetime.now() - startTyp, datetime.now()))
            
    if len(types) > 1:
        logging.info("Finished {} types at {} in {}".format(len(types), datetime.now(), datetime.now() - start))
        
# ############################# Driver ####################################

"""
Form of configs

    {
        "stationNumber": "999",
        "types": ["2"],
        "createDateByType": {"2": "date_entered_into_file" },
        "forceRedo": False
    }
    
"""
def main():

    if len(sys.argv) < 2:
        print "need to specify configuration ex/ {VISTAAB}.json - exiting"
        return

    configBase = sys.argv[1].split(".")[0]

    if not os.path.isfile("{}.json".format(configBase)):
        print "No config file {}.json - exiting".format(configBase)
        return

    config = json.load(open("{}.json".format(configBase)), object_pairs_hook=OrderedDict)

    if "stationNumber" not in config:
        print "No station number in configuration - exiting"
        return
    stationNumber = config["stationNumber"]
    
    dataLocnTempl=DATA_LOCN_TEMPL if "dataLocationTemplate" not in config else config["dataLocationTemplate"] # allows RF override etc. May nix just for that option?

    # "types": [], "createDateByType": {"{TYPE}": "{creation_date_property}" ...}
    # ... if don't set "types" then all types will be used
    if "types" not in config or len(config["types"]) == 0:
        dataLocn = dataLocnTempl.format(stationNumber)    
        print "Typing ALL available data as no subset specified"
        types = FMQLReplyStore(dataLocn).availableTypes()
    else:
        types = config["types"]
            
    reduceReportTypes(
        stationNumber, 
        types,
        dataLocnTempl=dataLocnTempl,
        forceRedo=(True if ("forceRedo" in config and config["forceRedo"] == True) else False),
        subTypeProps=(config["subTypeProps"] if "subTypeProps" in config else {}),
        createDate=({} if "createDate" not in config else config["createDate"]), 
        onAndAfterDay=(config["onAndAfterDay"] if "onAndAfterDay" in config else ""),
        upToDay=(config["upToDay"] if "upToDay" in config else ""),
        reductionLabel=(config["reductionLabel"] if "reductionLabel" in config else ""),
        countDateMonth=(True if "countDateMonth" in config and config["countDateMonth"] else False),
        countDateMonthFor=(config["countDateMonthFor"] if "countDateMonthFor" in config else []),
        forceCountProperties=(config["forceCountProperties"] if "forceCountProperties" in config else {}), 
        forceCountPointerTypesExtra=(config["forceCountPointerTypesExtra"] if "forceCountPointerTypesExtra" in config else {})
    )
        
if __name__ == "__main__":
    main()
    



#!/usr/bin/env python
# -*- coding: utf8 -*-

import os
import json
from datetime import datetime
import re

"""
TMP: Makes V2 Schema

- id is type of data ex/ PATIENT-2
- keep fmId for easy, unambig lookup (called fmId)
- location in properties as need (called fmLocation)
... note: all the fmX makes it easy to split out a DM without the fm
"""
def generateSchemaSummary(schemaLocn, tmpWorkingLocation):

    reducedDefns = []
    reducedDefnsByFMId = {}
    mpropsByChildFMId = {}
    RESERVED_PROPS = ["id", "_id", "type", "value", "parent", "label", "ien"] # to match new form (ex/ type in 19 qualified)
    for sfl in os.listdir(schemaLocn):
        if not re.match(r'SCHEMA', sfl):
            continue
        schemaJSON = json.load(open(schemaLocn + sfl))
        info = {"fmId": schemaJSON["number"], "properties": []}
        info["id"] = schemaJSON["name"] + "-" + re.sub(r'\.', '_', schemaJSON["number"])
        reducedDefnsByFMId[schemaJSON["number"]] = info
        if "parent" in schemaJSON:
            info["parentFMId"] = schemaJSON["parent"]
            if len(schemaJSON["fields"]) == 1:
                info["isSingleton"] = True
        if "description" in schemaJSON:
            info["description"] = schemaJSON["description"]["value"]
        for fieldInfo in schemaJSON["fields"]:
            if "location" not in fieldInfo:
                continue # computed prop
            pred = fieldInfo["pred"]
            # See that 44: "type2"/ 19: "type4" is captured and now flattened in transforms
            if pred in RESERVED_PROPS: # to match V2 qualification
                pred = "{}_{}".format(pred, re.sub(r'\.', '_', fieldInfo["number"]))
            if re.match(r'\d', pred): # applied in v2er too don't want numbers starting properties
                pred = "_" + pred
            propInfo = {"id": pred, "fmId": fieldInfo["number"], "fmLocation": fieldInfo["location"], "fmFlags": fieldInfo["flags"]} # want locn for grouping
            if pred != fieldInfo["pred"]:
                propInfo["origPred"] = fieldInfo["pred"]
            info["properties"].append(propInfo)
            # ****** fill in type for all - embed enum here that will use in FMQL
            propInfo["type"] = "TO BE FILLED IN"
            if fieldInfo["type"] == "9": # Multiple - only see fmId
                mpropsByChildFMId[fieldInfo["details"]] = fieldInfo["pred"]
                propInfo["type"] = "MULTIPLE" # need to spec 'multiple' below
                propInfo["fmChildId"] = fieldInfo["details"]
                continue
            # For Older FMQL too (as its data will have boolean
            if fieldInfo["type"] in ["3", "12"]: # boolean is codes too
                codesIE = [codeIE for codeIE in fieldInfo["details"].split(";") if codeIE != ""]
                propInfo["type"] = "CODES_BOOLEAN" if fieldInfo["type"] == "12" else "CODES"
                propInfo["range"] = codesIE
                continue
        reducedDefns.append(info)

    # Fill in holes in hierarchy - multiple gets parent's full id and property; parent gets multiple's id
    for info in reducedDefns:
        if "parentFMId" in info:
            pinfo = reducedDefnsByFMId[info["parentFMId"]]
            info["parent"] = pinfo["id"]
            info["parentProperty"] = mpropsByChildFMId[info["fmId"]] # want children to have parent property (do in FM schema too)
        for propInfo in info["properties"]:
            if "type" in propInfo and propInfo["type"] == "MULTIPLE":
                if propInfo["fmChildId"] not in reducedDefnsByFMId:
                    propInfo["bad"] = True
                    print "*** Saw bad property - multiple has invalid range: {}", propInfo
                else:
                    propInfo["childId"] = reducedDefnsByFMId[propInfo["fmChildId"]]["id"]

    json.dump(reducedDefns, open(tmpWorkingLocation + "schemaSummary.json", "w"))

    return reducedDefns 



#!/usr/bin/env python
# -*- coding: utf8 -*-

import sys
import os
import re
import json
from datetime import datetime
from collections import defaultdict, OrderedDict
import copy
import logging
import shutil

"""
From type data summary (transformers.TypeFromData), generate MongoDB supported JSON Schema definition

Outside package that checks schema: https://pypi.org/project/jsonschema/

** Sibling of upcoming jsonSchemaFromFMQLSchema **

Some Links:
- https://docs.mongodb.com/manual/core/schema-validation/
- https://docs.mongodb.com/manual/reference/operator/query/jsonSchema/#op._S_jsonSchema
- https://tools.ietf.org/html/draft-zyp-json-schema-04
- https://tools.ietf.org/html/draft-fge-json-schema-validation-00
- https://docs.mongodb.com/manual/reference/operator/query/type/#document-type-available-types
- https://docs.mongodb.com/manual/reference/operator/query/jsonSchema/
"""
def toJSONSchemaMongo(typeData):

    # Choice is to list forms (can have mix of local and national) and make IEN
    # permissive if not INT
    # ... was an ex of 120_51-SPO2%
    def makeIdRegExp(patterns):
        pMUs = []
        for p in patterns:
            if len(patterns[p]) > 1 or patterns[p].keys()[0] != "INT":
                ienMU = ".+" # int or completely permissive
            else:
                ienMU = "\d+"
            pMUs.append("{}:{}".format(re.sub(r'urn:', '', p), ienMU))
        return "urn:{}$".format(pMUs[0] if len(pMUs) == 1 else "({})".format("|".join(sorted(pMUs))))

    POINTER_DEFN = OrderedDict([("type", "object"), ("properties", OrderedDict([("id", OrderedDict([("type", "string"), ("pattern", "")])), ("label", {"type": "string"})])), ("required", ["id", "label"])])

    BASE_DEFN_PROPS = OrderedDict([("_id", OrderedDict([("type", "string"), ("pattern", makeIdRegExp(typeData["id"]))])), ("label", {"type": "string"}), ("type", {"type": "string"})])

    """
    Just do Objects first
    """
    def createObjectDefinition(count, propInfos, level=0, parentQProp="", problemPropInfos={}):

        def pointerPropDefn(propInfo):
            defn = copy.deepcopy(POINTER_DEFN)
            defn["properties"]["id"]["pattern"] = makeIdRegExp(propInfo["pointerPatterns"])
            return defn

        properties = OrderedDict() if parentQProp else BASE_DEFN_PROPS
        # https://docs.mongodb.com/manual/reference/operator/query/jsonSchema/#op._S_jsonSchema (interplay bsonType/type)
        simpleType = {
            "STRING": "string",
            "INTEGER": "number", # or int (32 bit) for bson
            "BOOLEAN": "boolean" # bool for bson
        }
        # Note: for VAM, need to add localStationNumber? iff local (ie/ not nationals)
        required = [] if parentQProp else ["_id", "label", "type"]
        for qprop in propInfos:
            noSl = len(qprop.split("/")) - 1
            if noSl != level:
                continue
            if parentQProp and not re.match(parentQProp, qprop):
                continue
            propInfo = propInfos[qprop]
            if "allTypesSeen" in propInfo:
                problemPropInfos[qprop] = propInfo
                continue
            prop = qprop.split("/")[-1]
            pcount = propInfo["count"]
            if pcount == count:
                required.append(prop)
            if propInfo["type"] in simpleType:
                defn = {"type": simpleType[propInfo["type"]]}
                properties[prop] = defn
                continue
            # Only basic or simple type that has to be bsontype
            if propInfo["type"] == "DATETIME":
                defn = {"bsonType": "date"}
                properties[prop] = defn
                continue
            if propInfo["type"] == "POINTER":
                properties[prop] = pointerPropDefn(propInfo)
                continue
            if propInfo["type"] == "ENUM":
                defn = {"enum": sorted([enumValue for enumValue in propInfo["enumValues"]])}
                properties[prop] = defn
                continue
            if propInfo["type"] == "OBJECT":
                properties[prop] = createObjectDefinition(pcount, propInfos, level+1, qprop, problemPropInfos=problemPropInfos)
                continue
            if re.match(r'\[', propInfo["type"]):
                if re.search(r'OBJECT', propInfo["type"]):
                    items = createObjectDefinition(propInfo["elementCount"], propInfos, level+1, qprop, problemPropInfos=problemPropInfos)
                elif re.search(r'ENUM', propInfo["type"]):
                    items = {"enum": sorted([enumValue for enumValue in propInfo["enumValues"]])}
                elif re.search(r'POINTER', propInfo["type"]):
                    items = pointerPropDefn(propInfo)
                else:
                    items = {"type": simpleType[re.search(r'\[([^\]]+)\]', propInfo["type"]).group(1)]}
                properties[prop] = OrderedDict([("type", "array"), ("items", items)])

        defn = OrderedDict([("type", "object"), ("properties", properties), ("additionalProperties", False)])
        if len(required):
            defn["required"] = required
        return defn

    totalRecords = typeData["count"]
    # Can't magically make booleans into strings where not already. Schema could become permissive but values would still vary. Need to correct.
    problemPropInfos = {}
    defn = createObjectDefinition(typeData["count"], typeData["properties"], problemPropInfos=problemPropInfos)
    if len(problemPropInfos):
        print "Can't generate Schema as ambiguous type for one or more properties:"
        for qprop in problemPropInfos:
            print "{} - {}. Can't do schema".format(qprop, json.dumps(problemPropInfos[qprop]))
        raise Exception("Stopping as {} problem properties".format(len(problemPropInfos)))
    return defn

